//======================================================== file = arbit.c =====
//=  Arbitrators for VOQ switch simulation model                              =
//=============================================================================
//=  Notes: 1)                                                                =
//=---------------------------------------------------------------------------=
//=  Contains: arbit_pim() - Implements PIM arbitration                       =
//=            arbit_islip() - Implements iSLIP arbitration                   =
//=            arbit_ppboq() - Implements PP-VOQ arbitration                  =
//=---------------------------------------------------------------------------=
//=  History:  KJC/KY (02/20/01) - Genesis (from voq_1c 1.02)                 =
//=============================================================================
#define EXTERNAL        // This is an external file (used in common.h)

//----- Include files ---------------------------------------------------------
#include "csim.h"       // Needed for CSIM
#include "common.h"     // Needed for common stuff

//=================================================================== PIM =====
//= Implements PIM arbitration                                                =
//=============================================================================
void arbit_pim(void)
{
  int      request[N][N];    // Request array (IP to OP)
  int      grant[N][N];      // Grant array (OP to IP)
  int      accept[N][N];     // Accept array (IP to OP)
  int      op_commit[N];     // TRUE iff OP i is committed (matched)
  int      ip_commit[N];     // TRUE iff IP i is committed (matched)
  int      count;            // Generic counter
  int      match;            // Generic match value
  int      ip, op;           // Counters for IP and OP
  int      iter;             // Counter for iterations

  create("arbit_pim");

  // Arbitrate the ports forever
  while(1)
  {
    // Clear accept[][], ip_commit[], and op_commit[]
    for (ip=0; ip<N; ip++)
      for (op=0; op<N; op++)
        accept[ip][op] = ip_commit[ip] = op_commit[op] = FALSE;

    // Do for NUM_ITER
    for (iter=0; iter<NUM_ITER; iter++)
    {
      // Load request array from Q_event[][] for un-matched IP's
      for (ip=0; ip<N; ip++)
        for (op=0; op<N; op++)
          if ((queue_cnt(Q_event[ip][op]) > 0) && (ip_commit[ip] == FALSE))
            request[ip][op] = TRUE;
          else
            request[ip][op] = FALSE;

      // Clear grant[][]
      for (ip=0; ip<N; ip++)
        for (op=0; op<N; op++)
          grant[ip][op] = FALSE;

      // For all OP's do grants to IP's
      for (op=0; op<N; op++)
      {
        // Check if this OP is committed (matched)
        if (op_commit[op] == TRUE) continue;

        // Count the number of requests
        count = 0;
        for (ip=0; ip<N; ip++)
          if (request[ip][op] == TRUE) count++;

        // Set grant[][] according to random grant
        if (count > 0)
        {
          match = random_int(1, count);
          match = (match == 0) ? 1 : match;

          for (ip=0; ip<N; ip++)
            if (request[ip][op] == TRUE)
            {
              match--;
              if (match == 0)
              {
                grant[ip][op] = TRUE;
                break;
              }
            }
         }
      }

      // For all IP's do accepts to OP's
      for (ip=0; ip<N; ip++)
      {
        // Count the number of grants
        count = 0;
        for (op=0; op<N; op++)
          if (grant[ip][op] == TRUE) count++;

        // Set accept[] according to random grant, clear matched request[][]
        // and set commit[]
        if (count > 0)
        {
          match = random_int(1, count);
          match = (match == 0) ? 1 : match;

          for (op=0; op<N; op++)
            if (grant[ip][op] == TRUE)
            {
              match--;
              if (match == 0)
              {
                accept[ip][op] = TRUE;
                ip_commit[ip] = TRUE;
                op_commit[op] = TRUE;
                request[ip][op] = FALSE;
                break;
              }
            }
        }
      }
    }

    // Fire the cells for all accept[][]
    for (ip=0; ip<N; ip++)
      for (op=0; op<N; op++)
        if (accept[ip][op] == TRUE)
        {
          set(Inport_busy[ip]);
          set(Outport_busy[op]);
          set(Q_event[ip][op]);
        }

    // Hold for a cell time and then an empty hold for synchronization
    hold((8.0 * CELL_SIZE)/(SPEED_UP * MEDIA_RATE));
    hold(0.0);
  }

  return;
}

//================================================================= ISLIP =====
//= Implements iSLIP arbitration                                              =
//=============================================================================
void arbit_islip(void)
{
  int      request[N][N];    // Request array (IP to OP)
  int      grant[N][N];      // Grant array (OP to IP)
  int      accept[N][N];     // Accept array (IP to OP)
  int      op_commit[N];     // TRUE iff OP i is committed (matched)
  int      ip_commit[N];     // TRUE iff IP i is committed (matched)
  int      next_ip, next_op; // Next values for ip and op
  int      ip, op;           // Counters for IP and OP
  int      g[N], a[N];       // iSLIP "g" and "a" pointers
  int      iter;             // Counter for iterations

  create("arbit_islip");

  // Initialize g[] and a[]
  for (ip=0; ip<N; ip++)
    for (op=0; op<N; op++)
      g[op] = a[ip] = 0;

  // Arbitrate the ports forever
  while(1)
  {
    // Clear accept[][], ip_commit[], and op_commit[]
    for (ip=0; ip<N; ip++)
      for (op=0; op<N; op++)
        accept[ip][op] = ip_commit[ip] = op_commit[op] = FALSE;

    // Do for NUM_ITER
    for (iter=0; iter<NUM_ITER; iter++)
    {
      // Load request array from Q_event[][] for un-matched IP's
      for (ip=0; ip<N; ip++)
        for (op=0; op<N; op++)
          if ((queue_cnt(Q_event[ip][op]) > 0) && (ip_commit[ip] == FALSE))
            request[ip][op] = TRUE;
          else
            request[ip][op] = FALSE;

      // Clear grant[][]
      for (ip=0; ip<N; ip++)
        for (op=0; op<N; op++)
          grant[ip][op] = FALSE;

      // For all OP's do grants to IP's
      for (op=0; op<N; op++)
      {
        // Check if this OP is committed (matched)
        if (op_commit[op] == TRUE) continue;

        // Set grant[][] according to iSLIP
        next_ip = g[op];
        for (ip=0; ip<N; ip++)
        {
          if (request[next_ip][op] == TRUE)
          {
            grant[next_ip][op] = TRUE;
            break;
          }
          next_ip = (next_ip + 1) % N;
        }
      }

      // For all IP's do accepts to OP's
      for (ip=0; ip<N; ip++)
      {
        // Set accept[] according to iSLIP
        next_op = a[ip];
        for (op=0; op<N; op++)
        {
          if (grant[ip][next_op] == TRUE)
          {
            accept[ip][next_op] = TRUE;
            ip_commit[ip] = TRUE;
            op_commit[next_op] = TRUE;
            request[ip][next_op] = FALSE;
            if (iter == 0)
            {
              a[ip] = (next_op + 1) % N;
              g[next_op] = (ip + 1) % N;
            }
            break;
          }
          next_op = (next_op + 1) % N;
        }
      }
    }

    // Fire the cells for all accept[][]
    for (ip=0; ip<N; ip++)
      for (op=0; op<N; op++)
        if (accept[ip][op] == TRUE)
        {
          set(Inport_busy[ip]);
          set(Outport_busy[op]);
          set(Q_event[ip][op]);
        }

    // Hold for a cell time and then an empty hold for synchronization
    hold((8.0 * CELL_SIZE)/(SPEED_UP * MEDIA_RATE));
    hold(0.0);
  }

  return;
}

//================================================================= PPVOQ =====
//= Implements PP-VOQ arbitration                                             =
//=============================================================================
void arbit_ppvoq(void)
{
  int      ip_token[N];      // IP value that OP token i is residing at
  int      ip_match[N];      // IP value that OP i is matched to
  int      op_match[N];      // OP value that IP i is matched to
  int      op_match_last[N]; // OP value that IP i matched to last time
  int      ip, op;           // Counters for IP and OP
  int      i, j;             // Generic loop counters

  create("arbit_ppvoq");

  // Initialize all OP tokens and OP last matched to IP 0
  for (i=0; i<N; i++)
    ip_token[i] = op_match_last[i] = 0;

  // Poll the ports forever
  while(1)
  {
    // Clear all previous matches (negative value marks a port as unmatched)
    for (i=0; i<N; i++)
      ip_match[i] = op_match[i] = -1;

    // Do one complete scan
    for (i=0; i<N; i++)
    {
      // Rotate all unmatched tokens one IP
      for (op=0; op<N; op++)
        if (ip_match[op] < 0) ip_token[op] = (ip_token[op] + 1) % N;

      // Check for and make matches in each IP
      for (ip=0; ip<N; ip++)
      {
        // Check if this IP is already matched
        if (op_match[ip] < 0)
        {
          // Rotate through all VOQ's in an unmatched IP checking if token
          // is present and VOQ has packets queued.  Start at the next
          // VOQ after the last one matched in this IP.
          op = op_match_last[ip];
          for (j=0; j<N; j++)
          {
            op = (op + 1) % N;
            if ((ip_token[op] == ip) && (queue_cnt(Q_event[ip][op]) > 0))
            {
              ip_match[op] = ip;
              op_match[ip] = op;
              break;
            }
          }
          if (op_match[ip] >= 0) op_match_last[ip] = op_match[ip];
        }
      }
    }

    // Fire the cells for all IP-OP matches
    for (op=0; op<N; op++)
      if (ip_match[op] >= 0)
      {
        set(Inport_busy[ip_match[op]]);
        set(Outport_busy[op]);
        set(Q_event[ip_match[op]][op]);
      }

    // Hold for a cell time and then an empty hold for synchronization
    hold((8.0 * CELL_SIZE)/(SPEED_UP * MEDIA_RATE));
    hold(0.0);
  }

  return;
}
