//========================================================= file = main.c =====
//=  Main program for a VOQ switch simulation model                           =
//=============================================================================
//=  Notes:  1) Simulation output is written to created file OUTFILE defined  =
//=             in common.h                                                   =
//=          2) Implements PIM, iSLIP, and PPVOQ arbitration.                 =
//=---------------------------------------------------------------------------=
//=  Build:  Build as a CSIM application linking to appropriate CSIM libs.    =
//=          Component files are:                                             =
//=            - common.h, main.c, generate.c, inport.c, outport.c, arbit.c   =
//=---------------------------------------------------------------------------=
//=  History:  KJC/KY (02/20/01) - Genesis (from voq_1c 1.02)                 =
//=============================================================================
#define MAIN                    // This is main (used in common.h)

//----- Include files ---------------------------------------------------------
#include <stdio.h>              // Needed for printf() and fprintf()
#include <stdlib.h>             // Needed for exit()
#include <math.h>               // Needed for sqrt()
#include "csim.h"               // Needed for CSIM
#include "common.h"             // Needed for common stuff

//=============================================================================
//==  Main program where sim() is the first process for a CSIM simulation    ==
//=============================================================================
void sim(void)
{
  FILE     *fp;                 // File pointer for output file
  int      run_num;             // Simulation run number
  double   global_mean;         // Overall (global) mean delay
  double   global_std_dev;      // Overall (global) std dev delay
  int      i, j;                // Generic loop counters

  // Set model name for CSIM
  set_model_name("VOQ packet switch simulation model");

  // Increase maximum number of allowed processes, facilities, and servers
  max_facilities(1000L);
  max_servers(10000L);
  max_processes(1000000L);
  max_events(1000000L);

  // Open output file for writing (delete and re-write if file already exists)
  fp = fopen(OUT_FILE, "w");
  if (fp == NULL)
  {
    printf("  >>> ERROR in creating output file '%s' \n", OUT_FILE);
    exit(1);
  }

  // Output banner to stdout and output file
  printf("*** BEGIN SIMULATION -- Version %s *** \n", VERSION);
  fprintf(fp,"*** BEGIN SIMULATION -- Version %s *** \n", VERSION);

  // *** Main simulation loop ***
  for (run_num=9; run_num<10; run_num++)
  {
    create("sim");

    // Initialize run parameters
    Sim_time = 10.0;
    Lambda1 = (run_num / 10.0) * 2118.2850;

    // Assign pointers to generators, inports, segmenters, outports,
    // and arbitrators
    for (i=0; i<N; i++)
    {
      if (GEN_FLAG == 1) gen[i] = gen_poisson;
      if (GEN_FLAG == 2) gen[i] = gen_ipp;
      if (GEN_FLAG == 3) gen[i] = gen_cbr;
      if (GEN_FLAG == 4) gen[i] = gen_square;
      seg[i] = segmenter;
      in[i] = inport;
      out[i] = outport;
    }

    // Variable initializations
    for (i=0; i<N; i++)
      Inport_count[i] = Outport_count[i] = 0;

    // CSIM initializations
    for (i=0; i<N; i++)
    {
      Outport_buf[i] = facility("Outport buffer for reassembled packets");
      Bytes_table[i] = table("Bytes transfered table");
      Delay_table[i] = table("Packet delay table");
      Inport_busy[i] = event("Event for inport is busy");
      Outport_busy[i] = event("Event for outport is busy");
      Xmit_done[i] = event("Event for transmit done at outport");
      for (j=0; j<N; j++)
        Q_event[i][j] = event("Event for queueing at inport i for outport j");
    }

    // Kick-off the generators
    for (i=0; i<N; i++)
      gen[i](i);

    // Kick-off the arbitrator after a one bit delay
    hold(1.0 / MEDIA_RATE);
    if (ARBIT_TYPE == PIM) arbit_pim();
    if (ARBIT_TYPE == ISLIP) arbit_islip();
    if (ARBIT_TYPE == PPVOQ) arbit_ppvoq();

    // Run until Sim_time expires while outputting status messages to stdout
    run_sim(Sim_time, run_num);

    // Compute global statistics
    global_mean = global_std_dev = 0.0;
    for (i=0; i<N; i++)
    {
      global_mean = global_mean + table_mean(Delay_table[i]) / N;
      global_std_dev = global_std_dev + sqrt(table_var(Delay_table[i])) / N;
    }

    // Output simulation configuration parameters
    fprintf(fp,"&*****************************************************\n");
    fprintf(fp,"&** Results from VOQ packet switch simulation model **\n");
    if (ARBIT_TYPE == PIM)
      fprintf(fp,"&**            >>> PIM arbitation <<<               **\n");
    if (ARBIT_TYPE == ISLIP)
      fprintf(fp,"&**           >>> iSLIP arbitation <<<              **\n");
    if (ARBIT_TYPE == PPVOQ)
      fprintf(fp,"&**          >>> PP-VOQ arbitation <<<              **\n");
    fprintf(fp,"&*****************************************************\n");
    if (GEN_FLAG == 1)
    {
      fprintf(fp,"&**   >>> Poisson with IEEE workgroup mix           \n");
      fprintf(fp,"&** Arrival rate        = %5.2f pkts/ms    \n",
        Lambda1);
    }
    if (GEN_FLAG == 2)
      fprintf(fp,"&**   >>> IPP with fixed packet length               \n");
    if (GEN_FLAG == 3)
      fprintf(fp,"&**   >>> CBR with fixed packet length               \n");
    if (GEN_FLAG == 4)
      fprintf(fp,"&**   >>> Square wave with fixed packet length       \n");
    fprintf(fp,"&** Simulation time      = %7.0f ms                    \n",
      Sim_time);
    fprintf(fp,"&** Media rate           = %9.0f Mbps                  \n",
      0.001 * MEDIA_RATE);
    fprintf(fp,"&** Internal speed-up    = %d                          \n",
      SPEED_UP);
    fprintf(fp,"&** Switch size          = %d ports                    \n",
      N);
    if ((ARBIT_TYPE == PIM) || (ARBIT_TYPE == ISLIP))
      fprintf(fp,"&** Iterations           = %d iterations             \n",
        NUM_ITER);

    // Output per port statistics
    for (i=0; i<N; i++)
    {
      fprintf(fp,"%2d====================================================\n",
        i);
      fprintf(fp,"%2d= Outport utilization  = %5.2f %%       \n", i,
        100.0 * (8.0 * table_sum(Bytes_table[i])) / (MEDIA_RATE * Sim_time));
      fprintf(fp,"%2d= Mean delay           = %f ms          \n", i,
        table_mean(Delay_table[i]));
      fprintf(fp,"%2d= Std Dev delay        = %f ms          \n", i,
        sqrt(table_var(Delay_table[i])));
    }

    // Output global statistics
    fprintf(fp,"&&====================================================\n");
    fprintf(fp,"&&  global mean delay    = %f ms \n", global_mean);
    fprintf(fp,"&&  global std dev delay = %f ms \n", global_std_dev);
    fprintf(fp,"&&====================================================\n");

    // Reset everything for a possible rerun
    rerun();
    reset_prob(1L);
  }

  // Close and end
  printf("*** END OF SIMULATION *** \n");
  fprintf(fp,"*** END OF SIMULATION *** \n");
  fclose(fp);
}
